// A. Sheaff 02/24/05

// Image class - provides basic image routines for gray scale images.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include "image.hh"

// From width and Height
Image::Image(int wd, int ht)
{
	w=wd;
	h=ht;
	d=new Pixel[wd*ht];
	memset(d,0,wd*ht);
	return;
}

// From width and Height and initial color
Image::Image(int wd, int ht, Pixel co)
{
	int i,j;

	w=wd;
	h=ht;
	d=new Pixel[wd*ht];
	for (i=0;i<ht;i++) {
		for (j=0;j<wd;j++) {
			d[index(j,i)]=co;
		}
	}
	return;
}

// From another image
Image::Image(const Image& i)
{
	w=i.width();
	h=i.height();
	d=new Pixel[w*h];
	memcpy(d,i.data(),w*h);
	return;
}

// From a file
Image::Image(const char * filename, const int wd, const int ht) : d(NULL)
{
	int fd, bytes;
	int sum=0;
	char buf[256];

	w=wd;
	h=ht;
	if (d) {
		delete[] d;
	}
	d=new Pixel[w*h];
	fd=open(filename,O_RDONLY);
	if (fd<0) {
		sprintf(buf,"open: %s",filename);
		perror(buf);
		memset(d,0,w*h);
		return;
	}

	sum=read(fd,d,wd*ht);
	if (sum<0) {
		sprintf(buf,"read: %s",filename);
		perror(buf);
		memset(d,0,w*h);
	}
	while (sum<wd*ht) {
		bytes=read(fd,&d[sum],wd*ht-sum);
		sum+=bytes;
		//printf("%d %d\n",sum,bytes);
		if (bytes<0) {
			sprintf(buf,"read: %s",filename);
			perror(buf);
			memset(d,0,w*h);
			break;
		}
	}
	close(fd);
	return;
}

// Destory an image
Image::~Image(void)
{
	w=0;
	h=0;
	if (d) {
		delete[] d;
	}
	d=NULL;
}

// Get a pixel from an image
Pixel Image::getpixel(const int x, const int y) const
{
	
	return d[index(x,y)];
}

// Set a pixel in an image
Pixel Image::setpixel(const int x, const int y, const Pixel p)
{
	Pixel pt;
	pt=(Pixel)d[index(x,y)];
	d[index(x,y)]=p;
	return pt;
}

// Helper: Calculate the index in the data array
inline int Image::index(const int x, const int y) const
{
	int i=y;
	int j=x;

	if (x<0) {
		j=0;
	}
	if (y<0) {
		i=0;
	}
	if (x>=w) {
		j=w-1;
	}
	if (y>=h) {
		i=h-1;
	}
	return (w*i+j);
}

// Size of an image
int Image::size(void)
{
	
	return (w*h);
}

// Copy an image
Image& Image::operator=(const Image& i)
{
	if (this!=&i) {
		if (d) {
			delete[] d;
		}
		w=i.width();
		h=i.height();
		d=new Pixel[w*h];
		memcpy(d,i.data(),w*h);
	}
	return (*this);
}

// Invert an image
Image Image::operator~(void)
{
	Image ti(*this);
	int i,j;	// Loop counters
	Pixel * dat1=ti.data();

	// Non-MMX.  For each pixel, invert and mask.
	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			dat1[index(j,i)]=(~d[index(j,i)])&0x00ffffff;
		}
	}
	
	return (ti);
}

// Add a constant to an image
Image& Image::operator+=(const Pixel c)
{
	int i,j;

	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			if (d[index(j,i)]+c>255) {
				d[index(j,i)]=255;
			} else {
				if (d[index(j,i)]+c<0) {
					d[index(j,i)]=0;
				} else {
					d[index(j,i)]+=c;
				}
			}
		}
	}
	return (*this);
}

// subtract a constant to an image
Image& Image::operator-=(const Pixel c)
{
	int i,j;

	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			if (d[index(j,i)]-c>255) {
				d[index(j,i)]=255;
			} else {
				if (d[index(j,i)]-c<0) {
					d[index(j,i)]=0;
				} else {
					d[index(j,i)]-=c;
				}
			}
		}
	}
	return (*this);
}

// or a constant to an image
Image& Image::operator|=(const Pixel c)
{
	int i,j;

	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			d[index(j,i)]|=c;
		}
	}
	return (*this);
}

// and a constant to an image
Image& Image::operator&=(const Pixel c)
{
	int i,j;

	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			d[index(j,i)]&=c;
		}
	}
	return (*this);
}

// xor a constant to an image
Image& Image::operator^=(const Pixel c)
{
	int i,j;

	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			d[index(j,i)]^=c;
		}
	}
	return (*this);
}

int Image::threshold(Pixel lt, Pixel ut)
{
	int i,j;

	for (i=0;i<h;i++) {
		for (j=0;j<w;j++) {
			if ((d[w*i+j]>lt)&&(d[w*i+j]<ut)) {
				d[w*i+j]=255;
			} else {
				d[w*i+j]=0;
			}
		}
	}
	return 0;
}

void Image::display(void)
{
	xdi.paint((unsigned char *)d,w,h);
	return;
}

void Image::refresh(void)
{
	xdi.refresh();
	return;
}

void Image::pause(void)
{
	xdi.waitforevent();
	return;
}

void Image::title(char * t)
{
	xdi.title(t);
	return;
}

void Image::drawline(int x1, int y1, int x2, int y2)
{
	xdi.drawline(x1, y1, x2, y2);
	return;
}

void Image::setlinewidth(int w)
{
	xdi.setlinewidth(w);
	return;
}

void Image::setcolor(unsigned long c)
{
	xdi.setcolor(c);
	return;
}

void Image::drawtext(int x, int y, char * string)
{
	xdi.drawtext(x, y, string);
	return;
}

// Write image to a file
int Image::writeimage(char * filename)
{
	int fd, err;
	char buf[356];

	fd=creat(filename,0644);
	if (fd<0) {
		sprintf(buf,"open: %s",filename);
		perror(buf);
		return -errno;
	}
	err=write(fd,d,w*h);
	if (err<0) {
		sprintf(buf,"write: %s",filename);
		perror(buf);
		return -errno;
	}
	close(fd);
	return(err);
}
